#!/usr/bin/env python
# Copyright (C) 2013, Xcoo, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-

import re
try:
    from hashlib import md5
except ImportError:
    from md5 import md5
from hashhtml import HashHtml
from header import Header
from line import Line
from list import List
from code import Code
from spangamut import SpanGamut


class BlockGamut(object):
    """Block Gamut Class"""
    def __init__(self, ws_re):
        self._ws_only_line_re = ws_re

    def reset(self, header, line, code, hashHtml, spangamut):
        self._header = header
        self._line = line
        self._code = code
        self._hash_html = hashHtml
        self._spangamut = spangamut
        self.extras = hashHtml.extras

    def run_block_gamut_with_list(self, text, list):
        """
        These are all the transformations that form block-level
        tags like paragraphs, headers, and list items.
        """
        text = self._header.do_headers(text)
        text = self._line.do_line(text)
        text = list.form_lists(text)
        if "pyshell" in self.extras:
            text = self._prepare_pyshell_blocks(text)
        if "wiki-tables" in self.extras:
            text = self._do_wiki_tables(text)
        text = self._code.do_code_blocks(text)
        text = self._do_block_quotes(text)
        # We already ran _HashHTMLBlocks() before, in Markdown(), but that
        # was to escape raw HTML in the original Markdown source. This time,
        # we're escaping the markup we've just created, so that we don't wrap
        # <p> tags around block-level tags.
        text = self._hash_html.hash_html_blocks(text)
        text = self._form_paragraphs(text)
        return text

    def _pyshell_block_sub(self, match):
        lines = match.group(0).splitlines(0)
        self._dedentlines(lines)
        indent = ' ' * self.tab_width
        s = ('\n'  # separate from possible cuddled paragraph
                + indent + ('\n' + indent).join(lines)
                + '\n\n')
        return s
    # Recipe: dedent (0.1.2)

    def _dedentlines(self, lines, tabsize=8, skip_first_line=False):
        """_dedentlines(lines, tabsize=8, skip_first_line=False) -> dedented lines

            "lines" is a list of lines to dedent.
            "tabsize" is the tab width to use for indent width calculations.
            "skip_first_line" is a boolean indicating if the first line should
            be skipped for calculating the indent width and for dedenting.
            This is sometimes useful for docstrings and similar.

            Same as dedent() except operates on a sequence of lines. Note: the
            lines list is modified **in-place**.
            """
        DEBUG = False
        if DEBUG:
            print "dedent: dedent(..., tabsize=%d, skip_first_line=%r)"\
                % (tabsize, skip_first_line)
        #    indents = []
        margin = None
        for i, line in enumerate(lines):
            if i == 0 and skip_first_line: continue
            indent = 0
            for ch in line:
                if ch == ' ':
                    indent += 1
                elif ch == '\t':
                    indent += tabsize - (indent % tabsize)
                elif ch in '\r\n':
                    continue  # skip all-whitespace lines
                else:
                    break
            else:
                continue  # skip all-whitespace lines
            if DEBUG: print "dedent: indent=%d: %r" % (indent, line)
            if margin is None:
                margin = indent
            else:
                margin = min(margin, indent)
        if DEBUG: print "dedent: margin=%r" % margin

        if margin is not None and margin > 0:
            for i, line in enumerate(lines):
                if i == 0 and skip_first_line: continue
                removed = 0
                for j, ch in enumerate(line):
                    if ch == ' ':
                        removed += 1
                    elif ch == '\t':
                        removed += tabsize - (removed % tabsize)
                    elif ch in '\r\n':
                        if DEBUG: print "dedent: %r: EOL -> strip up to EOL" % line
                        lines[i] = lines[i][j:]
                        break
                    else:
                        raise ValueError("unexpected non-whitespace char %r in "
                                         "line %r while removing %d-space margin"
                                         % (ch, line, margin))
                    if DEBUG:
                        print "dedent: %r: %r -> removed %d/%d"\
                            % (line, ch, removed, margin)
                    if removed == margin:
                        lines[i] = lines[i][j + 1:]
                        break
                    elif removed > margin:
                        lines[i] = ' ' * (removed - margin) + lines[i][j + 1:]
                        break
                else:
                    if removed:
                        lines[i] = lines[i][removed:]
        return lines

    def _prepare_pyshell_blocks(self, text):
        """Ensure that Python interactive shell sessions are put in
            code blocks -- even if not properly indented.
            """
        if ">>>" not in text:
            return text
        less_than_tab = self.tab_width - 1
        _pyshell_block_re = re.compile(r"""
            ^([ ]{0,%d})>>>[ ].*\n   # first line
            ^(\1.*\S+.*\n)*         # any number of subsequent lines
            ^\n                     # ends with a blank line
            """ % less_than_tab, re.M | re.X)
        return _pyshell_block_re.sub(self._pyshell_block_sub, text)

    _block_quote_re = re.compile(r'''
        (                           # Wrap whole match in \1
        (
        ^[ \t]*>[ \t]?          # '>' at the start of a line
        .+\n                  # rest of the first line
        (.+\n)*                 # subsequent consecutive lines
        \n*                     # blanks
        )+
        )
        ''', re.M | re.X)

    _bq_one_level_re = re.compile('^[ \t]*>[ \t]?', re.M)
    _html_pre_block_re = re.compile(r'(\s*<pre>.+?</pre>)', re.S)

    def _dedent_two_spaces_sub(self, match):
        return re.sub(r'(?m)^  ', '', match.group(1))

    def _block_quote_sub(self, match):
        bq = match.group(1)
        bq = self._bq_one_level_re.sub('', bq)  # trim one level of quoting
        bq = self._ws_only_line_re.sub('', bq)  # trim whitespace-only lines
        bq = self._run_block_gamut(bq)          # recurse
        bq = re.sub('(?m)^', '  ', bq)
        # These leading spaces screw with <pre> content, so we need to fix that:
        bq = self._html_pre_block_re.sub(self._dedent_two_spaces_sub, bq)
        return "<quote>\n%s\n</quote>\n\n" % bq

    def _do_block_quotes(self, text):
        if '>' not in text:
            return text
        return self._block_quote_re.sub(self._block_quote_sub, text)

    def _wiki_table_sub(self, match):
        ttext = match.group(0).strip()
        #print 'wiki table: %r' % match.group(0)
        rows = []
        for line in ttext.splitlines(0):
            line = line.strip()[2:-2].strip()
            row = [c.strip() for c in re.split(r'(?<!\\)\|\|', line)]
            rows.append(row)
        #pprint(rows)
        hlines = ['<table>', '<tbody>']
        for row in rows:
            hrow = ['<tr>']
            for cell in row:
                hrow.append('<td>')
                hrow.append(self._spangamut.run_span_gamut(cell))
                hrow.append('</td>')
            hrow.append('</tr>')
            hlines.append(''.join(hrow))
        hlines += ['</tbody>', '</table>']
        return '\n'.join(hlines) + '\n'

    def _do_wiki_tables(self, text):
        # Optimization.
        if "||" not in text:
            return text

        less_than_tab = self.tab_width - 1
        wiki_table_re = re.compile(r'''
            (?:(?<=\n\n)|\A\n?)            # leading blank line
            ^([ ]{0,%d})\|\|.+?\|\|[ ]*\n  # first line
            (^\1\|\|.+?\|\|\n)*        # any number of subsequent lines
            ''' % less_than_tab, re.M | re.X)
        return wiki_table_re.sub(self._wiki_table_sub, text)

    def _form_paragraphs(self, text):
        # Strip leading and trailing lines:
        text = text.strip('\n')
        # Wrap <p> tags.
        grafs = []
        for i, graf in enumerate(re.split(r"\n{2,}", text)):
            if graf in self._hash_html.get_htmlblocks_hash():
                # Unhashify HTML blocks
                html_bl = self._hash_html.get_htmlblocks_hash()
                grafs.append(html_bl[graf])
            else:
                cuddled_list = None
                if "cuddled-lists" in self.extras:
                    # Need to put back trailing '\n' for `_list_item_re`
                    # match at the end of the paragraph.
                    li = self._list_item_re.search(graf + '\n')
                    # Two of the same list marker in this paragraph: a likely
                    # candidate for a list cuddled to preceding paragraph
                    # text (issue 33). Note the `[-1]` is a quick way to
                    # consider numeric bullets (e.g. "1." and "2.") to be
                    # equal.
                    if (li and len(li.group(2)) <= 3 and li.group("next_marker")
                        and li.group("marker")[-1] == li.group("next_marker")[-1]):
                        start = li.start()
                        cuddled_list = self._list.do_lists(graf[start:]).rstrip("\n")
                        assert cuddled_list.startswith("<list")
                        graf = graf[:start]

                # Wrap <p> tags.
                graf = self._spangamut.run_span_gamut(graf)
                grafs.append("<p>" + graf.lstrip(" \t") + "</p>")

                if cuddled_list:
                    grafs.append(cuddled_list)
        return "\n\n".join(grafs)
