#!/usr/bin/env python
# Copyright (C) 2013, Xcoo, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-

import re
try:
    from hashlib import md5
except ImportError:
    from md5 import md5


class Code(object):
    '''Codeblock Class'''
    def reset(self, extras, tab_width, escape_table):
        self.tab_width = tab_width
        self.extras = extras
        self._escape_table = escape_table
        self._outdent_re = re.compile(r'^(\t|[ ]{1,%d})' % tab_width, re.M)
        # Cribbed from a post by Bart Lateur:
        # <http://www.nntp.perl.org/group/perl.macperl.anyperl/154>
        self._detab_re = re.compile(r'(.*?)\t', re.M)

    def _get_pygments_lexer(self, lexer_name):
        try:
            from pygments import lexers, util
        except ImportError:
            return None
        try:
            return lexers.get_lexer_by_name(lexer_name)
        except util.ClassNotFound:
            return None

    def _color_with_pygments(self, codeblock, lexer, **formatter_opts):
        import pygments
        import pygments.formatters

        class HtmlCodeFormatter(pygments.formatters.HtmlFormatter):
            def _wrap_code(self, inner):
                """A function for use in a Pygments Formatter which
                    wraps in <codeblock> tags.
                    """
                yield 0, "<codeblock>"
                for tup in inner:
                    yield tup
                yield 0, "</codeblock>"

            def wrap(self, source, outfile):
                """Return the source with a code, pre, and div."""
                return self._wrap_div(self._wrap_pre(self._wrap_code(source)))

        formatter = HtmlCodeFormatter(cssclass="codehilite", **formatter_opts)
        return pygments.highlight(codeblock, lexer, formatter)

    def _code_block_sub(self, match):
        codeblock = match.group(1)
        codeblock = self._outdent(codeblock)
        codeblock = self.detab(codeblock)
        #print codeblock
        codeblock = codeblock.lstrip('\n')  # trim leading newlines
        codeblock = codeblock.rstrip()      # trim trailing whitespace

        if "code-color" in self.extras and codeblock.startswith(":::"):
            lexer_name, rest = codeblock.split('\n', 1)
            lexer_name = lexer_name[3:].strip()
            lexer = self._get_pygments_lexer(lexer_name)
            codeblock = rest.lstrip("\n")   # Remove lexer declaration line.
            if lexer:
                formatter_opts = self.extras['code-color'] or {}
                colored = self._color_with_pygments(codeblock, lexer,
                                                    **formatter_opts)
                return "\n\n%s\n\n" % colored

        codeblock = self._encode_code(codeblock)
        pre_class_str = self._html_class_str_from_tag("pre")
        code_class_str = self._html_class_str_from_tag("code")
        return "\n\n<pre%s><codeblock%s>%s\n</codeblock></pre>\n\n" % (
                  pre_class_str, code_class_str, codeblock.replace('-->', '--&gt;').replace('<!--', '&lt;!--'))

    def _html_class_str_from_tag(self, tag):
        """Get the appropriate ' class="..."' string (note the leading
            space), if any, for the given tag.
            """
        if "html-classes" not in self.extras:
            return ""
        try:
            html_classes_from_tag = self.extras["html-classes"]
        except TypeError:
            return ""
        else:
            if tag in html_classes_from_tag:
                return ' class="%s"' % html_classes_from_tag[tag]
        return ""
    
    def do_code_blocks_for_git(self, text):
        """Process Markdown '<codeblock class="git"> blocks."""
        code_block_re = re.compile(r'''
            (?:\n\n|\A)(^```(.*?)\n(.*?)(?:^```|\Z))
            ''', re.S | re.M | re.X)
        result = code_block_re.findall(text)
        if len(result) > 0:
            for i in range(len(result)):
                code = '<pre><codeblock language="%s" class="git">%s</codeblock></pre>' % (result[i][1], result[i][2].replace('>', '&gt;').replace('<', '&lt;'))
        
                text = text.replace(result[i][0], code)
        return text

    def do_code_blocks(self, text):
        """Process Markdown `<pre><codeblock>` blocks."""
        code_block_re = re.compile(r'''
            (?:\n\n|\A)
            ( # $1 = the code block -- one or more lines, starting with a space/tab
            (?:
            (?:[ ]{%d} | \t)  # Lines must start with a tab or a tab-width of spaces
            .*\n+
            )+
            )
            ((?=^[ ]{0,%d}\S)|\Z)   # Lookahead for non-space at line-start, or end of doc
            ''' % (self.tab_width, self.tab_width),
                                   re.M | re.X)

        return code_block_re.sub(self._code_block_sub, text)

    def _outdent(self, text):
        # Remove one level of line-leading tabs or spaces
        return self._outdent_re.sub('', text)

    def _detab_sub(self, match):
        g1 = match.group(1)
        return g1 + (' ' * (self.tab_width - len(g1) % self.tab_width))

    def detab(self, text):
        r"""Remove (leading?) tabs from a file.

            >>> m = Markdown()
            >>> m._detab("\tfoo")
            '    foo'
            >>> m._detab("  \tfoo")
            '    foo'
            >>> m._detab("\t  foo")
            '      foo'
            >>> m._detab("  foo")
            '  foo'
            >>> m._detab("  foo\n\tbar\tblam")
            '  foo\n    bar blam'
            """
        if '\t' not in text:
            return text
        return self._detab_re.subn(self._detab_sub, text)[0]

    # Rules for a code span:
    # - backslash escapes are not interpreted in a code span
    # - to include one or or a run of more backticks the delimiters must
    #   be a longer run of backticks
    # - cannot start or end a code span with a backtick; pad with a
    #   space and that space will be removed in the emitted HTML
    # See `test/tm-cases/escapes.text` for a number of edge-case
    # examples.
    _code_span_re = re.compile(r'''
        (?<!\\)
        (`+)        # \1 = Opening run of `
        (?!`)       # See Note A test/tm-cases/escapes.text
        (.+?)       # \2 = The code block
        (?<!`)
        \1          # Matching closer
        (?!`)
        ''', re.X | re.S)
    
    def _code_span_sub(self, match):
        c = match.group(2).strip(" \t")
        c = self._encode_code(c)
        return "<codeblock>%s</codeblock>" % c
            
    def do_code_spans(self, text):
        ''' Backtick quotes are used for <code></code> spans.'''
        #
        #   *   You can use multiple backticks as the delimiters if you want to
        #       include literal backticks in the code span. So, this input:
        #
        #         Just type ``foo `bar` baz`` at the prompt.
        #
        #       Will translate to:
        #
        #         <p>Just type <code>foo `bar` baz</code> at the prompt.</p>
        #
        #       There's no arbitrary limit to the number of backticks you
        #       can use as delimters. If you need three consecutive backticks
        #       in your code, use four for delimiters, etc.
        #
        #   *   You can use spaces to get literal backticks at the edges:
        #
        #         ... type `` `bar` `` ...
        #
        #       Turns to:
        #
        #         ... type <code>`bar`</code> ...
        return self._code_span_re.sub(self._code_span_sub, text)

    def _encode_code(self, text):
        """Encode/escape certain characters inside Markdown code runs.
            The point is that in code, these characters are literals,
            and lose their special Markdown meanings.
            """
        replacements = [
            # Encode all ampersands; HTML entities are not
            # entities within a Markdown code span.
                        #('&', '&amp;'),
            # Do the angle bracket song and dance:
                        #('<', '&lt;'),
                        # ('>', '&gt;'),
            # Now, escape characters that are magic in Markdown:
            ('*', self._escape_table['*']),
            ('_', self._escape_table['_']),
            ('{', self._escape_table['{']),
            ('}', self._escape_table['}']),
            ('[', self._escape_table['[']),
            (']', self._escape_table[']']),
            ('\\', self._escape_table['\\']),
            ]
        for before, after in replacements:
            text = text.replace(before, after)
        return text
