#!/usr/bin/env python
# Copyright (C) 2013, Xcoo, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-

import sys
import re
try:
    from hashlib import md5
except ImportError:
    from md5 import md5
from hashlist import HashList
from mderror import MarkdownError
from memorized import _memoized

#---- Python version compat

if sys.version_info[:2] < (2, 4):
    from sets import Set as set

    def reversed(sequence):
        for i in sequence[::-1]:
            yield i

    def _unicode_decode(s, encoding, errors='xmlcharrefreplace'):
        return unicode(s, encoding, errors)
else:

    def _unicode_decode(s, encoding, errors='strict'):
        return s.decode(encoding, errors)


class HashHtml(object):
    '''Hash table Class'''
    # I broke out the html5 tags here and add them to _block_tags_a and
    # _block_tags_b.  This way html5 tags are easy to keep track of.
    _html5tags = '|article|aside|header|hgroup|footer|nav|section|figure|figcaption'
    _block_tags_a = 'p|div|h[1-6]|blockquote|pre|table|dl|ol|ul|script|noscript|form|fieldset|iframe|math|ins|del|quote'
    _block_tags_a += _html5tags

    _strict_tag_block_re = re.compile(r"""
        (                       # save in \1
        ^                   # start of line  (with re.M)
        <(%s)               # start tag = \2
        \b                  # word break
        (.*\n)*?            # any number of lines, minimally matching
        </\2>               # the matching end tag
        [ \t]*              # trailing spaces/tabs
        (?=\n+|\Z)          # followed by a newline or end of document
        )
        """ % _block_tags_a,
                                      re.X | re.M)
    _block_tags_b = 'p|div|h[1-6]|blockquote|pre|table|dl|ol|ul|script|noscript|form|fieldset|iframe|math|quote|s|strike'
    _block_tags_b += _html5tags
    _liberal_tag_block_re = re.compile(r"""
        (                       # save in \1
        ^                   # start of line  (with re.M)
        <(%s)               # start tag = \2
        \b                  # word break
        (.*\n)*?            # any number of lines, minimally matching
        .*</\2>             # the matching end tag
        [ \t]*              # trailing spaces/tabs
        (?=\n+|\Z)          # followed by a newline or end of document
        )
        """ % _block_tags_b,
                                       re.X | re.M)
    _block_tags_c = 'chapter|part|section|subsection|subsubsection|subsubsubsection|list'
    _list_tag_block_re = re.compile(r"""
        (                       # save in \1
        ^                   # start of line  (with re.M)
        <(%s)            # start tag = \2
        \b                  # word break
        (.*\n)*?            # any number of lines, minimally matching
        .*</\2>             # the matching end tag
        [ \t]*              # trailing spaces/tabs
        (?=\n+\n|\n?\Z)          # followed by a newline or end of document
        )
        """ % _block_tags_c, re.X | re.M)

    _block_tags_d = 'chapter|part|section|subsection|subsubsection|subsubsubsection'
    _title_tag_block_re = re.compile(r"""
        (
        ^
        <(%s)[ \t]+title[ \t]*=[ \t]*".*?"[ \t]*>
        [ \t]*              # trailing spaces/tabs
        (?=\n+|\Z) 
        )
        """ % _block_tags_d,
                                     re.X | re.M)

    def reset(self, ex, tabwidth, safe):
        self._html_spans_hash = HashList()
        self._html_blocks_hash = HashList()
        self.extras = ex
        self.tab_width = tabwidth
        self.safe_mode = safe

    def get_htmlblocks_hash(self):
        return self._html_blocks_hash.get_hash_list()

    def hash_html_spans(self, text):
        # Used for safe_mode.

        def _is_auto_link(s):
            if ':' in s and self._auto_link_re.match(s):
                return True
            elif '@' in s and self._auto_email_link_re.match(s):
                return True
            return False

        tokens = []
        is_html_markup = False
        for token in self._sorta_html_tokenize_re.split(text):
            if is_html_markup and not _is_auto_link(token):
                sanitized = self._sanitize_html(token)
                key = self._html_spans_hash.create_hash_utf8(sanitized)
                tokens.append(key)
            else:
                tokens.append(token)
            is_html_markup = not is_html_markup
        return ''.join(tokens)

    def unhash_html_spans(self, text):
        for key, sanitized in self.html_spans.items():
            text = text.replace(key, sanitized)
        return text

    def _sanitize_html(self, s):
        if self.safe_mode == "replace":
            return self.html_removed_text
        elif self.safe_mode == "escape":
            replacements = [
                            ('&', '&amp;'),
                            ('<', '&lt;'),
                            ('>', '&gt;'),
                            ]
            for before, after in replacements:
                s = s.replace(before, after)
            return s
        else:
            raise MarkdownError("invalid value for 'safe_mode': %r (must be "
                                "'escape' or 'replace')" % self.safe_mode)

    def _hash_html_block_sub(self, match, raw=False):
        html = match.group(1)
        if raw and self.safe_mode:
            html = self._sanitize_html(html)
        elif 'markdown-in-html' in self.extras and 'markdown=' in html:
            first_line = html.split('\n', 1)[0]
            m = self._html_markdown_attr_re.search(first_line)
            if m:
                lines = html.split('\n')
                middle = '\n'.join(lines[1:-1])
                last_line = lines[-1]
                first_line = first_line[:m.start()] + first_line[m.end():]
                f_key = self._html_blocks_hash.create_hash_utf8(first_line)
                l_key = self._html_blocks_hash.create_hash_utf8(last_line)
                return ''.join(["\n\n", f_key,
                                "\n\n", middle, "\n\n",
                                l_key, "\n\n"])

        key = self._html_blocks_hash.create_hash_utf8(html)
        return "\n\n" + key + "\n\n"

    def hash_html_blocks(self, text, raw=False):
        """Hashify HTML blocks

            We only want to do this for block-level HTML tags, such as headers,
            lists, and tables. That's because we still want to wrap <p>s around
            "paragraphs" that are wrapped in non-block-level tags, such as anchors,
            phrase emphasis, and spans. The list of tags we're looking for is
            hard-coded.

            @param raw {boolean} indicates if these are raw HTML blocks in
            the original source. It makes a difference in "safe" mode.
            """

        if '<' not in text:
            return text
        # Pass `raw` value into our calls to self._hash_html_block_sub.
        hash_html_block_sub = _curry(self._hash_html_block_sub, raw=raw)

        # First, look for nested blocks, e.g.:
        #   <div>
        #       <div>
        #       tags for inner block must be indented.
        #       </div>
        #   </div>
        #
        # The outermost tags must start at the left margin for this to match, and
        # the inner nested divs must be indented.
        # We need to do this before the next, more liberal match, because the next
        # match will start at the first `<div>` and stop at the first `</div>`.

        # added for list 
        text = self._list_tag_block_re.sub(hash_html_block_sub, text)

        text = self._strict_tag_block_re.sub(hash_html_block_sub, text)

        # Now match more liberally, simply from `\n<tag>` to `</tag>\n`
        text = self._liberal_tag_block_re.sub(hash_html_block_sub, text)

        # added for title
        text = self._title_tag_block_re.sub(hash_html_block_sub, text)

        # Special case just for <hr />. It was easier to make a special
        # case than to make the other regex more complicated.
        if "<line" in text:
            _hr_tag_re = _hr_tag_re_from_tab_width(self.tab_width)
            text = _hr_tag_re.sub(hash_html_block_sub, text)

        # Special case for standalone HTML comments:
        if "<!--" in text:
            start = 0
            while True:
                # Delimiters for next comment block.
                try:
                    start_idx = text.index("<!--", start)
                except ValueError, ex:
                    break
                try:
                    end_idx = text.index("-->", start_idx) + 3
                except ValueError, ex:
                    break
                # Start position for next comment block search.
                start = end_idx
                # Validate whitespace before comment.
                if start_idx:
                    # - Up to `tab_width - 1` spaces before start_idx.
                    for i in range(self.tab_width - 1):
                        if text[start_idx - 1] != ' ':
                            break
                        start_idx -= 1
                        if start_idx == 0:
                            break
                    # - Must be preceded by 2 newlines or hit the start of
                    #   the document.
                    if start_idx == 0:
                        pass
                    elif start_idx == 1 and text[0] == '\n':
                        start_idx = 0  # to match minute detail of Markdown.pl regex
                    elif text[start_idx - 2:start_idx] == '\n\n':
                        pass
                    else:
                        break
                # Validate whitespace after comment.
                # - Any number of spaces and tabs.
                while end_idx < len(text):
                    if text[end_idx] not in ' \t':
                        break
                    end_idx += 1
                # - Must be following by 2 newlines or hit end of text.
                if text[end_idx:end_idx + 2] not in ('', '\n', '\n\n'):
                    continue
                # Escape and hash (must match `_hash_html_block_sub`).
                html = text[start_idx:end_idx]
                if raw and self.safe_mode:
                    html = self._sanitize_html(html)
                # key = _hash_text(html)
                #self.html_blocks[key] = html
                key = self._html_blocks_hash.create_hash_utf8(html)
                text = text[:start_idx] + "\n\n" + key + "\n\n" + text[end_idx:]
        if "xml" in self.extras:
            # Treat XML processing instructions and namespaced one-liner
            # tags as if they were block HTML tags. E.g., if standalone
            # (i.e. are their own paragraph), the following do not get
            # wrapped in a <p> tag:
            #    <?foo bar?>
            #
            #    <xi:include xmlns:xi="http://www.w3.org/2001/XInclude" href="chapter_1.md"/>
            _xml_oneliner_re = _xml_oneliner_re_from_tab_width(self.tab_width)
            text = _xml_oneliner_re.sub(hash_html_block_sub, text)

        return text


# From http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52549
def _curry(*args, **kwargs):
    function, args = args[0], args[1:]

    def result(*rest, **kwrest):
        combined = kwargs.copy()
        combined.update(kwrest)
        return function(*args + rest, **combined)
    return result


def _hr_tag_re_from_tab_width(tab_width):
    return re.compile(r"""
        (?:
        (?<=\n\n)       # Starting after a blank line
        |               # or
        \A\n?           # the beginning of the doc
        )
        (                       # save in \1
        [ ]{0,%d}
        <(line)               # start tag = \2
        \b                  # word break
        ([^<>])*?           #
        /?>                 # the matching end tag
        [ \t]*
        (?=\n{2,}|\Z)       # followed by a blank line or end of document
        )
        """ % (tab_width - 1), re.X)
_hr_tag_re_from_tab_width = _memoized(_hr_tag_re_from_tab_width)
