#!/usr/bin/env python
# Copyright (C) 2013, Xcoo, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-

import re
try:
    from hashlib import md5
except ImportError:
    from md5 import md5
from spangamut import SpanGamut


class Header(object):

        _headerNames = ["part", "chapter", "section", "subsection", "subsubsection", "subsubsubsection"]

        def __init__(self, empty_element):
            self.empty_element_suffix = empty_element

        def reset(self, ex, spangamut):
            self.extras = ex
            self._spangamut = spangamut

        def _get_title_re(self, num):
            prenum = num - 1
            if prenum < 0:
                prenum = 0
            result_title = re.compile(r'''
                (^<(%s)\b(.*\n)*?(?=(?:%s\Z%s)))
                ''' % (self._headerNames[num], self._get_start_of_titles(num), self._get_end_of_pre_titles(prenum)), re.X | re.M)
            return result_title

        def _get_start_of_titles(self, num):
            result = ""
            while num >= 0:
                result += "<" + self._headerNames[num] + "|"
                num -= 1
            return result

        def _get_end_of_pre_titles(self, num):
            result = ""
            while num >= 0:
                result += "|</" + self._headerNames[num]
                num -= 1
            return result

        def _setext_h_end_sub(self, match):
            titletag = match.group(2)
            titlecontent = match.group(1)
            return "%s</%s>\n" % (titlecontent, titletag)

        def form_title(self, text):
            for cnt in range(len(self._headerNames)):
                setend_h_re = self._get_title_re(cnt)
                text = setend_h_re.sub(self._setext_h_end_sub, text)
            return text
        _toc = None

        def _toc_add_entry(self, level, id, name):
            if self._toc is None:
                self._toc = []
            self._toc.append((level, id, name))
        _setext_h_re = re.compile(r'^(.+)[ \t]*\n(=+|-+)[ \t]*\n+', re.M)

        def _setext_h_sub(self, match):
            n = {"=": 1, "-": 2}[match.group(2)[0]]
            demote_headers = self.extras.get("demote-headers")
            if demote_headers:
                n = min(n + demote_headers, 6)
            header_id_attr = ""
            if "header-ids" in self.extras:
                header_id = self.header_id_from_text(match.group(1),
                                                     self.extras["header-ids"], n)
                if header_id:
                    header_id_attr = ' id="%s"' % header_id
            html = self._spangamut.run_span_gamut(match.group(1))
            if "toc" in self.extras and header_id:
                self._toc_add_entry(n, header_id, html)
            return "<%s title = \"%s\">\n\n" % (self._headerNames[n - 1], html)

        _atx_h_re = re.compile(r'''
            ^(\#{1,6})  # \1 = string of #'s
            [ \t]*
            (.+?)       # \2 = Header text
            [ \t]*
            (?<!\\)     # ensure not an escaped trailing '#'
            \#*         # optional closing #'s (not counted)
            \n+
            ''', re.X | re.M)

        def _atx_h_sub(self, match):
            n = len(match.group(1))
            demote_headers = self.extras.get("demote-headers")
            if demote_headers:
                n = min(n + demote_headers, 6)
            header_id_attr = ""
            if "header-ids" in self.extras:
                header_id = self.header_id_from_text(match.group(2),
                                                     self.extras["header-ids"], n)
                if header_id:
                    header_id_attr = ' id="%s"' % header_id
            #TODO
            #html = self._spangamut.run_span_gamut(match.group(2))
            #html = match.group(2)
            title_re = re.compile('''"''')
            html = title_re.sub("&quot;",match.group(2))
            if "toc" in self.extras and header_id:
                self._toc_add_entry(n, header_id, html)
            return "\n<%s title = \"%s\">\n\n" % (self._headerNames[n - 1], html)

        # insert header's end
        #TODO  _atx_h_end_re = re.complile()

        def do_headers(self, text):
            # Setext-style headers:
            #     Header 1
            #     ========
            #
            #     Header 2
            #     --------
            text = self._setext_h_re.sub(self._setext_h_sub, text)

            # atx-style headers:
            #   # Header 1
            #   ## Header 2
            #   ## Header 2 with closing hashes ##
            #   ...
            #   ###### Header 6
            text = self._atx_h_re.sub(self._atx_h_sub, text)

            return text

        def header_id_from_text(self, text, prefix, n):
            """Generate a header id attribute value from the given header
                HTML content.

                This is only called if the "header-ids" extra is enabled.
                Subclasses may override this for different header ids.

                @param text {str} The text of the header tag
                @param prefix {str} The requested prefix for header ids. This is the
                value of the "header-ids" extra key, if any. Otherwise, None.
                @param n {int} The <hN> tag number, i.e. `1` for an <h1> tag.
                @returns {str} The value for the header tag's "id" attribute. Return
                None to not have an id attribute and to exclude this header from
                the TOC (if the "toc" extra is specified).
                """
            header_id = _slugify(text)
            if prefix and isinstance(prefix, basestring):
                header_id = prefix + '-' + header_id
            if header_id in self._count_from_header_id:
                self._count_from_header_id[header_id] += 1
                header_id += '-%s' % self._count_from_header_id[header_id]
            else:
                self._count_from_header_id[header_id] = 1
            return header_id


## {{{ http://code.activestate.com/recipes/577257/ (r1)
_slugify_strip_re = re.compile(r'[^\w\s-]')
_slugify_hyphenate_re = re.compile(r'[-\s]+')


def _slugify(value):
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.

    From Django's "django/template/defaultfilters.py".
    """
    import unicodedata
    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore')
    value = unicode(_slugify_strip_re.sub('', value).strip().lower())
    return _slugify_hyphenate_re.sub('-', value)
## end of http://code.activestate.com/recipes/577257/ }}}
