#!/usr/bin/env python
# Copyright (C) 2013, Xcoo, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-

import re
try:
    from hashlib import md5
except ImportError:
    from md5 import md5
from hashlist import HashList


class Links(object):
    urls = None
    titles = None

    def __init__(self, ex, table, ees, tab_width):
        self.extras = ex
        self._escape_table = table
        self.empty_element_suffix = ees
        self.tab_width = tab_width
        self.urls = {}
        self.titles = {}

    _tail_of_inline_link_re = re.compile(r'''
            # Match tail of: [text](/url/) or [text](/url/ "title")
            \(            # literal paren
            [ \t]*
            (?P<url>            # \1
            <.*?>
            |
            .*?
            )
            [ \t]*
            (                   # \2
            (['"])            # quote char = \3
            (?P<title>.*?)
            \3                # matching quote
            )?                  # title is optional
            \)
            ''', re.X | re.S)

    _tail_of_reference_link_re = re.compile(r'''
        # Match tail of: [text][id]
        [ ]?          # one optional space
        (?:\n[ ]*)?   # one optional newline followed by spaces
        \[
        (?P<id>.*?)
        \]
        ''', re.X | re.S)

    def do_links(self, text):
        """Turn Markdown link shortcuts into XHTML <a> and <img> tags.

            This is a combination of Markdown.pl's _DoAnchors() and
            _DoImages(). They are done together because that simplified the
            approach. It was necessary to use a different approach than
            Markdown.pl because of the lack of atomic matching support in
            Python's regex engine used in $g_nested_brackets.
            """
        MAX_LINK_TEXT_SENTINEL = 3000  # markdown2 issue 24
        # `anchor_allowed_pos` is used to support img links inside
        # anchors, but not anchors inside anchors. An anchor's start
        # pos must be `>= anchor_allowed_pos`.
        anchor_allowed_pos = 0

        curr_pos = 0
        while True: # Handle the next link.
            # The next '[' is the start of:
            # - an inline anchor:   [text](url "title")
            # - a reference anchor: [text][id]
            # - an inline img:      ![text](url "title")
            # - a reference img:    ![text][id]
            # - a footnote ref:     [^id]
            #   (Only if 'footnotes' extra enabled)
            # - a footnote defn:    [^id]: ...
            #   (Only if 'footnotes' extra enabled) These have already
            #   been stripped in _strip_footnote_definitions() so no
            #   need to watch for them.
            # - a link definition:  [id]: url "title"
            #   These have already been stripped in
            #   _strip_link_definitions() so no need to watch for them.
            # - not markup:         [...anything else...
            try:
                start_idx = text.index('[', curr_pos)
            except ValueError:
                break
            text_length = len(text)

            # Find the matching closing ']'.
            # Markdown.pl allows *matching* brackets in link text so we
            # will here too. Markdown.pl *doesn't* currently allow
            # matching brackets in img alt text -- we'll differ in that
            # regard.
            bracket_depth = 0
            for p in range(start_idx + 1, min(start_idx + MAX_LINK_TEXT_SENTINEL,
                                            text_length)):
                ch = text[p]
                if ch == ']':
                    bracket_depth -= 1
                    if bracket_depth < 0:
                        break
                elif ch == '[':
                    bracket_depth += 1
            else:
                # Closing bracket not found within sentinel length.
                # This isn't markup.
                curr_pos = start_idx + 1
                continue
            link_text = text[start_idx + 1:p]

            # Possibly a footnote ref?
            if "footnotes" in self.extras and link_text.startswith("^"):
                normed_id = re.sub(r'\W', '-', link_text[1:])
                if normed_id in self.footnotes:
                    self.footnote_ids.append(normed_id)
                    result = '<sup class="footnote-ref" id="fnref-%s">' \
                        '<link href="#fn-%s">%s</link></sup>' \
                        % (normed_id, normed_id, len(self.footnote_ids))
                    text = text[:start_idx] + result + text[p + 1:]
                else:
                    # This id isn't defined, leave the markup alone.
                    curr_pos = p + 1
                continue

            # Now determine what this is by the remainder.
            p += 1
            if p == text_length:
                return text

            # Inline anchor or img?
            if text[p] == '(':  # attempt at perf improvement
                match = self._tail_of_inline_link_re.match(text, p)
                if match:
                    # Handle an inline anchor or img.
                    is_img = start_idx > 0 and text[start_idx - 1] == "!"
                    if is_img:
                        start_idx -= 1

                    url, title = match.group("url"), match.group("title")
                    if url and url[0] == '<':
                        url = url[1:-1]  # '<url>' -> 'url'
                    # We've got to encode these to avoid conflicting
                    # with italics/bold.
                    url = url.replace('*', self._escape_table['*']) \
                        .replace('_', self._escape_table['_'])
                    if title:
                        title_str = ' title="%s"' % (
                                                     _xml_escape_attr(title)
                                                     .replace('*', self._escape_table['*'])
                                                     .replace('_', self._escape_table['_']))
                    else:
                        title_str = ''
                    if is_img:
                        result = '<image uri="%s" alt="%s"%s%s' \
                            % (url.replace('"', '&quot;'),
                               _xml_escape_attr(link_text),
                               title_str, self.empty_element_suffix)
                        if "smarty-pants" in self.extras:
                            result = result.replace('"', self._escape_table['"'])
                        curr_pos = start_idx + len(result)
                        text = text[:start_idx] + result + text[match.end():]
                    elif start_idx >= anchor_allowed_pos:
                        result_head = '<link uri="%s"%s>' % (url, title_str)
                        result = '%s%s</link>' % (result_head, link_text)
                        if "smarty-pants" in self.extras:
                            result = result.replace('"', self._escape_table['"'])
                        # <img> allowed from curr_pos on, <a> from
                        # anchor_allowed_pos on.
                        curr_pos = start_idx + len(result_head)
                        anchor_allowed_pos = start_idx + len(result)
                        text = text[:start_idx] + result + text[match.end():]
                    else:
                        # Anchor not allowed here.
                        curr_pos = start_idx + 1
                    continue

            # Reference anchor or img?
            else:
                match = self._tail_of_reference_link_re.match(text, p)
                if match:
                    # Handle a reference-style anchor or img.
                    is_img = start_idx > 0 and text[start_idx - 1] == "!"
                    if is_img:
                        start_idx -= 1
                    link_id = match.group("id").lower()
                    if not link_id:
                        link_id = link_text.lower()  # for links like [this][]
                    if link_id in self.urls:
                        url = self.urls[link_id]
                        # We've got to encode these to avoid conflicting
                        # with italics/bold.
                        url = url.replace('*', self._escape_table['*']) \
                            .replace('_', self._escape_table['_'])
                        title = self.titles.get(link_id)
                        if title:
                            before = title
                            title = _xml_escape_attr(title) \
                                .replace('*', self._escape_table['*']) \
                                .replace('_', self._escape_table['_'])
                            title_str = ' title="%s"' % title
                        else:
                            title_str = ''
                        if is_img:
                            result = '<image uri="%s" alt="%s"%s%s' \
                                % (url.replace('"', '&quot;'),
                                   link_text.replace('"', '&quot;'),
                                   title_str, self.empty_element_suffix)
                            if "smarty-pants" in self.extras:
                                result = result.replace('"', self._escape_table['"'])
                            curr_pos = start_idx + len(result)
                            text = text[:start_idx] + result + text[match.end():]
                        elif start_idx >= anchor_allowed_pos:
                            result = '<link uri="%s"%s>%s</link>' \
                                % (url, title_str, link_text)
                            result_head = '<link uri="%s"%s>' % (url, title_str)
                            result = '%s%s</link>' % (result_head, link_text)
                            if "smarty-pants" in self.extras:
                                result = result.replace('"', self._escape_table['"'])
                            # <img> allowed from curr_pos on, <a> from
                            # anchor_allowed_pos on.
                            curr_pos = start_idx + len(result_head)
                            anchor_allowed_pos = start_idx + len(result)
                            text = text[:start_idx] + result + text[match.end():]
                        else:
                            # Anchor not allowed here.
                            curr_pos = start_idx + 1
                    else:
                        # This id isn't defined, leave the markup alone.
                        curr_pos = match.end()
                    continue
            # Otherwise, it isn't markup.
            curr_pos = start_idx + 1
        return text

    _auto_link_re = re.compile(r'<((https?|ftp):[^\'">\s]+)>', re.I)
    def _auto_link_sub(self, match):
        g1 = match.group(1)
        return '<link uri="%s">%s</link>' % (g1, g1)

    _auto_email_link_re = re.compile(r"""
        <
        (?:mailto:)?
        (
        [-.\w]+
        \@
        [-\w]+(\.[-\w]+)*\.[a-z]+
        )
        >
        """, re.I | re.X | re.U)
    def _auto_email_link_sub(self, match):
        return self._encode_email_address(
                                          self._unescape_special_chars(match.group(1)))

    def do_auto_links(self, text):
        text = self._auto_link_re.sub(self._auto_link_sub, text)
        text = self._auto_email_link_re.sub(self._auto_email_link_sub, text)
        return text

    def _encode_email_address(self, addr):
        #  Input: an email address, e.g. "foo@example.com"
        #
        #  Output: the email address as a mailto link, with each character
        #      of the address encoded as either a decimal or hex entity, in
        #      the hopes of foiling most address harvesting spam bots. E.g.:
        #
        #    <a href="&#x6D;&#97;&#105;&#108;&#x74;&#111;:&#102;&#111;&#111;&#64;&#101;
        #       x&#x61;&#109;&#x70;&#108;&#x65;&#x2E;&#99;&#111;&#109;">&#102;&#111;&#111;
        #       &#64;&#101;x&#x61;&#109;&#x70;&#108;&#x65;&#x2E;&#99;&#111;&#109;</a>
        #
        #  Based on a filter by Matthew Wickline, posted to the BBEdit-Talk
        #  mailing list: <http://tinyurl.com/yu7ue>
        chars = [_xml_encode_email_char_at_random(ch)
                 for ch in "mailto:" + addr]
        # Strip the mailto: from the visible part.
        addr = '<link uri="%s">%s</link>' \
            % (''.join(chars), ''.join(chars[7:]))
        return addr

    def do_link_patterns(self, text):
        """Caveat emptor: there isn't much guarding against link
            patterns being formed inside other standard Markdown links, e.g.
            inside a [link def][like this].

            Dev Notes: *Could* consider prefixing regexes with a negative
            lookbehind assertion to attempt to guard against this.
            """
        # link_from_hash = {}
        link_from_hash = HashList()
        for regex, repl in self.link_patterns:
            replacements = []
            for match in regex.finditer(text):
                if hasattr(repl, "__call__"):
                    href = repl(match)
                else:
                    href = match.expand(repl)
                replacements.append((match.span(), href))
            for (start, end), href in reversed(replacements):
                escaped_href = (
                                href.replace('"', '&quot;')  # b/c of attr quote
                                # To avoid markdown <em> and <strong>:
                                .replace('*', self._escape_table['*'])
                                .replace('_', self._escape_table['_']))
                link = '<link uri="%s">%s</link>' % (escaped_href, text[start:end])
                hash = link_from_hash.create_hash_utf8(link)
                # hash = _hash_text(link)
                #link_from_hash[hash] = link
                text = text[:start] + hash + text[end:]
        text = link_from_hash.unhash(test)
        #for hash, link in link_from_hash.items():
        #text = text.replace(hash, link)
        return text
            
    def _strip_link_definitions(self, text):
        # Strips link definitions from text, stores the URLs and titles in
        # hash references.
        less_than_tab = self.tab_width - 1
        
        # Link defs are in the form:
        #   [id]: url "optional title"
        _link_def_re = re.compile(r"""
            ^[ ]{0,%d}\[(.+)\]: # id = \1
            [ \t]*
            \n?               # maybe *one* newline
            [ \t]*
            <?(.+?)>?           # url = \2
            [ \t]*
            (?:
            \n?             # maybe one newline
            [ \t]*
            (?<=\s)         # lookbehind for whitespace
            ['"(]
            ([^\n]*)        # title = \3
            ['")]
            [ \t]*
            )?  # title is optional
            (?:\n+|\Z)
            """ % less_than_tab, re.X | re.M | re.U)
        return _link_def_re.sub(self._extract_link_def_sub, text)
                    
    def _extract_link_def_sub(self, match):
        id, url, title = match.groups()
        key = id.lower()    # Link IDs are case-insensitive
        self.urls[key] = url#self._encode_amps_and_angles(url)
        if title:
            self.titles[key] = title
        return ""



def _xml_escape_attr(attr, skip_single_quote = True):
    """Escape the given string for use in an HTML/XML tag attribute.

        By default this doesn't bother with escaping `'` to `&#39;`, presuming that
        the tag attribute is surrounded by double quotes.
        """
    escaped = (attr
               .replace('&', '&amp;')
               .replace('"', '&quot;')
               .replace('<', '&lt;')
               .replace('>', '&gt;'))
    if not skip_single_quote:
        escaped = escaped.replace("'", "&#39;")
    return escaped
