#!/usr/bin/env python
# Copyright (C) 2013, Xcoo, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-

import re
try:
    from hashlib import md5
except ImportError:
    from md5 import md5
from hashlist import HashList
from spangamut import SpanGamut

class List(object):

    def __init__(self, empty_element):
        self.empty_element_suffix = empty_element

    def form_lists(self, text):
        text = self.do_lists(text)
        text = self._do_format_lists(text)
        return text

    def reset(self, tab_width, spangamut, list_level, blockgamut):
        self.tab_width = tab_width
        self.list_level = list_level
        self._item_hash = HashList()
        self._outdent_re = re.compile(r'^(\t|[ ]{1,%d})' % self.tab_width, re.M)
        self._span_gamut = spangamut
        self._block_gamut = blockgamut

    _marker_ul_chars  = '*+-'
    _marker_any = r'(?:[%s]|\d+\.)' % _marker_ul_chars
    _marker_ul = '(?:[%s])' % _marker_ul_chars
    _marker_ol = r'(?:\d+\.)'
    
    def _reform_list(self, text):
        #TODO modify regular expression
        add_item_list_re = re.compile(r'''
        ^<list''', re.X | re.M )
        text = add_item_list_re.sub("</item><list", text)
        return text

    def _list_sub(self, match):
        lst = match.group(1)
        lst_type = match.group(3) in self._marker_ul_chars and "normal" or "number"
        result = self._process_list_items(lst)
        result = self._reform_list(result)
        if self.list_level:
            return "<list type = \"%s\">\n%s</list>\n" % (lst_type, result)
        else:
            return "<list type = \"%s\">\n%s</list>\n\n" % (lst_type, result)

    def do_lists(self, text):
        # Form HTML ordered (numbered) and unordered (bulleted) lists.

        for marker_pat in (self._marker_ul, self._marker_ol):
            # Re-usable pattern to match any entire ul or ol list:
            less_than_tab = self.tab_width - 1
            whole_list = r'''
                (                   # \1 = whole list
                (                 # \2
                [ ]{0,%d}
                (%s)            # \3 = first list item marker
                [ \t]+
                (?!\ *\3\ )     # '- - - ...' isn't a list. See 'not_quite_a_list' test case.
                )
                (?:.+?)
                (                 # \4
                \Z
                |
                \n{2,}
                (?=\S)
                (?!           # Negative lookahead for another list item marker
                [ \t]*
                %s[ \t]+
                )
                )
                )
                ''' % (less_than_tab, marker_pat, marker_pat)

            if self.list_level:
                sub_list_re = re.compile("^"+whole_list, re.X | re.M | re.S)
                text = sub_list_re.sub(self._list_sub, text)
            else:
                list_re = re.compile(r"(?:(?<=\n\n)|\A\n?)"+whole_list,
                                     re.X | re.M | re.S)
                text = list_re.sub(self._list_sub, text)
        return text

    def _do_add_item(self, match):
        item_content = match.group(1)
        return "%s</item>" % item_content

    def _do_format_lists(self, text):
        sub_list_delete_item_re = re.compile(r'''
            </list></item>''', re.X | re.M )
        text = sub_list_delete_item_re.sub('</list>',text)
        sub_list_add_item_re = re.compile(r'''
            (^\S+(?=\n<list))''', re.X | re.M)
        text = sub_list_add_item_re.sub(self._do_add_item, text)
        sub_list_newline_re = re.compile(r'''
            (<item>(.*\n)*?.*?</item>)
            ''', re.X | re.M)
        text = sub_list_newline_re.sub(self._newline_to_space_sub, text)
        sub_list_none_item_re = re.compile(r'''
            </item>\s+</item>
            ''', re.X | re.M)
        text = sub_list_none_item_re.sub('''</item>''', text)
        return text

    _list_item_re = re.compile(r'''
        (\n)?                   # leading line = \1
        (^[ \t]*)               # leading whitespace = \2
        (?P<marker>%s) [ \t]+   # list marker = \3
        ((?:.+?)                # list item text = \4
        (\n{1,2}))             # eols = \5
        (?= \n* (\Z | \2 (?P<next_marker>%s) [ \t]+))
        ''' % (_marker_any, _marker_any),
                               re.M | re.X | re.S)
    _last_li_endswith_two_eols = False
    _item_re = re.compile(r'''(^[-*+])''', re.M | re.X | re.S)

    _list_p_re = re.compile(r'''^<p>(.+)</p>$''', re.M | re.X | re.S)
    _list_br_re = re.compile(r'''</p>\s+<p>''', re.M | re.X | re.S)
    
    def _add_br(self, match):
        item = match.group(1)
        item = self._list_br_re.sub("<br/>",item)
        return item
    
    def _add_newline(self, item):
        item = self._list_p_re.sub(self._add_br ,item)
        return item
    
    def _list_item_sub(self, match):
        item = match.group(4)
        item_header =  match.group(3)
        leading_line = match.group(1)
        leading_space = match.group(2)
        #key = _hash_text(item)
        key = self._item_hash.create_hash_utf8(item)

        if re.search(self._item_re, item_header):
            item_or_num = "item|"
        else:
            item_or_num = "number|"

        self._item_hash.set_hash_item(key, item_or_num)
        item = item_or_num + item

        if leading_line or "\n\n" in item or self._last_li_endswith_two_eols:
            item = self._block_gamut.run_block_gamut_with_list(self._outdent(item), self)
        else:
            # Recursion for sub-lists:
            item = self.do_lists(self._outdent(item))
            if item.endswith('\n'):
                item = item[:-1]
            item = self._span_gamut.run_span_gamut(item)
        item = self._add_newline(item)
        self._last_li_endswith_two_eols = (len(match.group(5)) == 2)
        return "<item>%s</item>\n" % item

    def _process_list_items(self, list_str):
        # Process the contents of a single ordered or unordered list,
        # splitting it into individual list items.
        # The $g_list_level global keeps track of when we're inside a list.
        # Each time we enter a list, we increment it; when we leave a list,
        # we decrement. If it's zero, we're not in a list anymore.
        #
        # We do this because when we're not inside a list, we want to treat
        # something like this:
        #
        #       I recommend upgrading to version
        #       8. Oops, now this line is treated
        #       as a sub-list.
        #
        # As a single paragraph, despite the fact that the second line starts
        # with a digit-period-space sequence.
        #
        # Whereas when we're inside a list (or sub-list), that line will be
        # treated as the start of a sub-list. What a kludge, huh? This is
        # an aspect of Markdown's syntax that's hard to parse perfectly
        # without resorting to mind-reading. Perhaps the solution is to
        # change the syntax rules such that sub-lists must start with a
        # starting cardinal number; e.g. "1." or "a.".
        self.list_level += 1
        self._last_li_endswith_two_eols = False
        list_str = list_str.rstrip('\n') + '\n'
        list_str = self._list_item_re.sub(self._list_item_sub, list_str)
        self.list_level -= 1
        return list_str

    def _outdent(self, text):
        # Remove one level of line-leading tabs or spaces
        return self._outdent_re.sub('', text)

    def _newline_to_space_sub(self, match):
        text = match.group(1)
        newline_re = re.compile(r'''\n[ \t]*
        ''', re.X | re.M)
        text = newline_re.sub(' ', text)
        return text
