#!/usr/bin/env python
# Copyright (C) 2013, Xcoo, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-

import sys
import re
import logging
try:
    from hashlib import md5
except ImportError:
    from md5 import md5
from unicodewithattrs import UnicodeWithAttrs
from hashhtml import HashHtml
from header import Header
from line import Line
from list import List
from code import Code
from spangamut import SpanGamut
from emacsvar import EmacsVars
from blockgamut import BlockGamut
from syntax import Syntax


#---- Python version compat

if sys.version_info[:2] < (2, 4):
    from sets import Set as set

    def reversed(sequence):
        for i in sequence[::-1]:
            yield i

    def _unicode_decode(s, encoding, errors='xmlcharrefreplace'):
        return unicode(s, encoding, errors)
else:

    def _unicode_decode(s, encoding, errors='strict'):
        return s.decode(encoding, errors)


#---- globals

DEBUG = False
log = logging.getLogger("markdown")

DEFAULT_TAB_WIDTH = 4


try:
    import uuid
except ImportError:
    SECRET_SALT = str(randint(0, 1000000))
else:
    SECRET_SALT = str(uuid.uuid4())


def _hash_ascii(s):
    #return md5(s).hexdigest()   # Markdown.pl effectively does this.
    return 'md5-' + md5(SECRET_SALT + s).hexdigest()


def _hash_text(s):
    return 'md5-' + md5(SECRET_SALT + s.encode("utf-8")).hexdigest()

# Table of hash values for escaped characters:
g_escape_table = dict([(ch, _hash_ascii(ch))
                       for ch in '\\`*_{}[]()>#+-.!'])


class Markdown(object):
    # The dict of "extras" to enable in processing -- a mapping of
    # extra name to argument for the extra. Most extras do not have an
    # argument, in which case the value is None.
    #
    # This can be set via (a) subclassing and (b) the constructor
    # "extras" argument.
    extras = None
    #item_blocks = None
    html_removed_text = "[HTML_REMOVED]"  # for compat with markdown.py
    # Used to track when we're inside an ordered or unordered list
    # (see _ProcessListItems() for details):
    list_level = 0

    _ws_only_line_re = re.compile(r"^[ \t]+$", re.M)

    def __init__(self, html4tags=False, tab_width=4, safe_mode=None,
                 extras=None, link_patterns=None, use_file_vars=False):
        if html4tags:
            self.empty_element_suffix = ">"
        else:
            self.empty_element_suffix = " />"
        self.tab_width = tab_width

        self._hash_html = HashHtml()
        self._header = Header(self.empty_element_suffix)
        self._line = Line(self.empty_element_suffix)
        self._list = List(self.empty_element_suffix)
        self._code = Code()
        self._blockgamut = BlockGamut(self._ws_only_line_re)
        self._syntax = Syntax()

        # For compatibility with earlier markdown2.py and with
        # markdown.py's safe_mode being a boolean,
        #   safe_mode == True -> "replace"
        if safe_mode is True:
            self.safe_mode = "replace"
        else:
            self.safe_mode = safe_mode

        # Massaging and building the "extras" info.
        if self.extras is None:
            self.extras = {}
        elif not isinstance(self.extras, dict):
            self.extras = dict([(e, None) for e in self.extras])
        if extras:
            if not isinstance(extras, dict):
                extras = dict([(e, None) for e in extras])
            self.extras.update(extras)
        assert isinstance(self.extras, dict)
        if "toc" in self.extras and not "header-ids" in self.extras:
            self.extras["header-ids"] = None   # "toc" implies "header-ids"
        self._instance_extras = self.extras.copy()

        self.link_patterns = link_patterns
        self.use_file_vars = use_file_vars
        self._outdent_re = re.compile(r'^(\t|[ ]{1,%d})' % tab_width, re.M)

        self._escape_table = g_escape_table.copy()
        if "smarty-pants" in self.extras:
            self._escape_table['"'] = _hash_ascii('"')
            self._escape_table["'"] = _hash_ascii("'")

    def reset(self):
        self.item_blocks = {}
        self.list_level = 0
        self.extras = self._instance_extras.copy()
        if "footnotes" in self.extras:
            self.footnotes = {}
            self.footnote_ids = []
        if "header-ids" in self.extras:
            self._count_from_header_id = {}  # no `defaultdict` in Python 2.4

        self._hash_html.reset(self.extras, self.tab_width, self.safe_mode)
        
        self._code.reset(self.extras, self.tab_width, self._escape_table)

        self._spangamut = SpanGamut(self._escape_table, self.extras, self.empty_element_suffix, self._code, self.tab_width)
        self._header.reset(self.extras, self._spangamut)
        self._blockgamut.reset(self._header, self._line, self._code, self._hash_html, self._spangamut)
        self._list.reset(self.tab_width, self._spangamut, self.list_level, self._blockgamut)
    
    def _count_amp(self, match):
        gts = match.group(1)
        gt_re = re.compile('''&gt;''', re.X | re.M)
        result = gt_re.sub('>',gts)
        return result
    
    def _convert_comment(self, match):
        comment = match.group(1)
        
        gt_re = re.compile('''&gt;''', re.X | re.M)
        result = gt_re.sub('>',comment)
        gt_re = re.compile('''&lt;''', re.X | re.M)
        result = gt_re.sub('<',result)
        return result

    def _convert_link(self, match):
        link = match.group(1)
        return "<%s>" % link

    def _convert_html(self, match):
        tag = match.group(1)
        content = match.group(2)
        content = self._html_blocks_re.sub(self._convert_html, content)
        return "<%s>%s</%s>" % (tag, content, tag)
            
    def _add_attribute(self, match):
        _code_re = re.compile('''<codeblock>\Z''',re.X | re.M)
        result = _code_re.sub('''<codeblock language="%s">''' % self._syntax.get_syntax(match.group(2)) ,match.group(1))
        return result
            
    def _add_code_attribute(self, text):
        _add_code_at_re = re.compile('''(<!--(.+)?-->\s*\n<pre>(<codeblock>))''',re.X | re.M)
        result = _add_code_at_re.sub(self._add_attribute,text)
        return result
    
    def convert(self, text):
        """Convert the given text."""
        # Main function. The order in which other subs are called here is
        # essential. Link and image substitutions need to happen before
        # _EscapeSpecialChars(), so that any *'s or _'s in the <a>
        # and <img> tags get encoded.

        # Clear the global hashes. If we don't clear these, you get conflicts
        # from other articles when generating a page which contains more than
        # one article (e.g. an index page that shows the N most recent
        # articles):
        self.reset()

        if not isinstance(text, unicode):
            #TODO: perhaps shouldn't presume UTF-8 for string input?
            text = unicode(text, 'utf-8')

        if self.use_file_vars:
            # Look for emacs-style file variable hints.
            emacs = EmacsVars()
            emacs_vars = self.emacs.get_emacs_vars(text)
            if "markdown-extras" in emacs_vars:
                splitter = re.compile("[ ,]+")
                for e in splitter.split(emacs_vars["markdown-extras"]):
                    if '=' in e:
                        ename, earg = e.split('=', 1)
                        try:
                            earg = int(earg)
                        except ValueError:
                            pass
                    else:
                        ename, earg = e, None
                    self.extras[ename] = earg

        # Standardize line endings:
        text = re.sub("\r\n|\r", "\n", text)

        # Make sure $text ends with a couple of newlines:
        text += "\n\n"

        # Convert all tabs to spaces.
        text = self._code.detab(text)
        
        
    


        # Strip any lines consisting only of spaces and tabs.
        # This makes subsequent regexen easier to write, because we can
        # match consecutive blank lines with /\n+/ instead of something
        # contorted like /[ \t]*\n+/ .
        text = self._ws_only_line_re.sub("", text)
        if self.safe_mode:
            text = self._hash_html.hash_html_spans(text)
                        
        
                        
        # Turn block-level HTML blocks into hash entries
        text = self._hash_html.hash_html_blocks(text, raw=True)

                # Convert all < and > and amp to ampasand.
        text = self._spangamut.convert_amps_and_angles(text)
        _quate_block_re = re.compile(r'''(^(&gt;)+)''', re.X | re.M)
        text = _quate_block_re.sub(self._count_amp, text)
        _comment_block_re = re.compile(r'''(&lt;!--[ \n]*((.+?)\n*)*?--&gt;)''', re.X | re.M)
        text = _comment_block_re.sub(self._convert_comment, text)
        _link_block_re = re.compile(r'''&lt;(https?://.*?)&gt;''', re.X | re.M)
        text = _link_block_re.sub(self._convert_link, text)
    
        # Strip link definitions, store in hashes.
        if "footnotes" in self.extras:
            # Must do footnotes first because an unlucky footnote defn
            # looks like a link defn:
            #   [^4]: this "looks like a link defn"
            text = self._strip_footnote_definitions(text)

        text = self._spangamut._links._strip_link_definitions(text)

       

        text = self._run_block_gamut(text)
            
        
        if "footnotes" in self.extras:
            text = self._add_footnotes(text)

        text = self.postprocess(text)

        text = self._unescape_special_chars(text)

        if self.safe_mode:
            text = self._hash_html.unhash_html_spans(text)
        text += "\n"
        #insert end of title
        text = self._header.form_title(text)
        #append document tag
        text = self._do_document(text)
        #delete excessive \n
        #text = self._delete_newline(text)
        #convert \n to space in paragraph
        text = self._convert_newline(text)
        #refactor <item><p> to <item>
        text = self._refactor_item(text)
        
        text = self._add_code_attribute(text)
            
        rv = UnicodeWithAttrs(text)
        rv = text
        if "toc" in self.extras:
            rv._toc = self._toc


        return rv
                
    def _refactor_item(self, text):
        itemp_re = re.compile(r'''<item><p>''', re.X | re.M)
        text = itemp_re.sub("<item>", text)
        pitem_re = re.compile(r'''</p></item>''', re.X | re.M)
        text = pitem_re.sub("</item>", text)
        return text

    def postprocess(self, text):
        """A hook for subclasses to do some postprocessing of the html, if
            desired. This is called before unescaping of special chars and
            unhashing of raw HTML spans.
            """
        return text

    # Cribbed from a post by Bart Lateur:
    # <http://www.nntp.perl.org/group/perl.macperl.anyperl/154>
    _detab_re = re.compile(r'(.*?)\t', re.M)

    def _detab_sub(self, match):
        g1 = match.group(1)
        return g1 + (' ' * (self.tab_width - len(g1) % self.tab_width))

    def _detab(self, text):
        r"""Remove (leading?) tabs from a file.

            >>> m = Markdown()
            >>> m._detab("\tfoo")
            '    foo'
            >>> m._detab("  \tfoo")
            '    foo'
            >>> m._detab("\t  foo")
            '      foo'
            >>> m._detab("  foo")
            '  foo'
            >>> m._detab("  foo\n\tbar\tblam")
            '  foo\n    bar blam'
            """
        if '\t' not in text:
            return text
        return self._detab_re.subn(self._detab_sub, text)[0]

    _html_markdown_attr_re = re.compile(
                                        r'''\s+markdown=("1"|'1')''')



    def _extract_footnote_def_sub(self, match):
        id, text = match.groups()
        text = _dedent(text, skip_first_line = not text.startswith('\n')).strip()
        normed_id = re.sub(r'\W', '-', id)
        # Ensure footnote text ends with a couple newlines (for some
        # block gamut matches).
        self.footnotes[normed_id] = text + "\n\n"
        return ""

    def _strip_footnote_definitions(self, text):
        """A footnote definition looks like this:
            [^note-id]: Text of the note.

            May include one or more indented paragraphs.

            Where,
            - The 'note-id' can be pretty much anything, though typically it
            is the number of the footnote.
            - The first paragraph may start on the next line, like so:

            [^note-id]:
            Text of the note.
            """
        less_than_tab = self.tab_width - 1
        footnote_def_re = re.compile(r'''
            ^[ ]{0,%d}\[\^(.+)\]:   # id = \1
            [ \t]*
            (                       # footnote text = \2
            # First line need not start with the spaces.
            (?:\s*.*\n+)
            (?:
            (?:[ ]{%d} | \t)  # Subsequent lines must be indented.
            .*\n+
            )*
            )
            # Lookahead for non-space at line-start, or end of doc.
            (?:(?=^[ ]{0,%d}\S)|\Z)
            ''' % (less_than_tab, self.tab_width, self.tab_width),
                                     re.X | re.M)
        return footnote_def_re.sub(self._extract_footnote_def_sub, text)

    def _run_block_gamut(self, text):
        # These are all the transformations that form block-level
        # tags like paragraphs, headers, and list items.
        text = self._header.do_headers(text)
        
        text = self._line.do_line(text)

        text = self._list.form_lists(text)


        if "pyshell" in self.extras:
            text = self._prepare_pyshell_blocks(text)
        if "wiki-tables" in self.extras:
            text = self._do_wiki_tables(text)
        
        text = self._code.do_code_blocks_for_git(text)
        text = self._hash_code(text)
        text = self._code.do_code_blocks(text)
        text = self._unhash_code(text)
                #        

        text = self._do_block_quotes(text)
        # We already ran _HashHTMLBlocks() before, in Markdown(), but that
        # was to escape raw HTML in the original Markdown source. This time,
        # we're escaping the markup we've just created, so that we don't wrap
        # <p> tags around block-level tags.

        text = self._hash_html.hash_html_blocks(text)
        
        text = self._form_paragraphs(text)
            
       
        
        return text

    def _pyshell_block_sub(self, match):
        lines = match.group(0).splitlines(0)
        _dedentlines(lines)
        indent = ' ' * self.tab_width
        s = ('\n'  # separate from possible cuddled paragraph
             + indent + ('\n' + indent).join(lines)
             + '\n\n')
        return s

    def _prepare_pyshell_blocks(self, text):
        """Ensure that Python interactive shell sessions are put in
            code blocks -- even if not properly indented.
            """
        if ">>>" not in text:
            return text

        less_than_tab = self.tab_width - 1
        _pyshell_block_re = re.compile(r"""
            ^([ ]{0,%d})>>>[ ].*\n   # first line
            ^(\1.*\S+.*\n)*         # any number of subsequent lines
            ^\n                     # ends with a blank line
            """ % less_than_tab, re.M | re.X)

        return _pyshell_block_re.sub(self._pyshell_block_sub, text)

    def _wiki_table_sub(self, match):
        ttext = match.group(0).strip()
        #print 'wiki table: %r' % match.group(0)
        rows = []
        for line in ttext.splitlines(0):
            line = line.strip()[2:-2].strip()
            row = [c.strip() for c in re.split(r'(?<!\\)\|\|', line)]
            rows.append(row)
        #pprint(rows)
        hlines = ['<table>', '<tbody>']
        for row in rows:
            hrow = ['<tr>']
            for cell in row:
                hrow.append('<td>')
                hrow.append(self._spangamut.run_span_gamut(cell))
                hrow.append('</td>')
            hrow.append('</tr>')
            hlines.append(''.join(hrow))
        hlines += ['</tbody>', '</table>']
        return '\n'.join(hlines) + '\n'

    def _do_wiki_tables(self, text):
        # Optimization.
        if "||" not in text:
            return text

        less_than_tab = self.tab_width - 1
        wiki_table_re = re.compile(r'''
            (?:(?<=\n\n)|\A\n?)            # leading blank line
            ^([ ]{0,%d})\|\|.+?\|\|[ ]*\n  # first line
            (^\1\|\|.+?\|\|\n)*        # any number of subsequent lines
            ''' % less_than_tab, re.M | re.X)
        return wiki_table_re.sub(self._wiki_table_sub, text)

    def _newline_to_space_sub(self, match):
        text = match.group(1)
        newline_re = re.compile(r'''\n[ \t]*
            ''', re.X | re.M)
        text = newline_re.sub(' ', text)
        return text

    def _convert_newline(self, text):
        text = self._hash_code(text)
        para_re = re.compile(r'''
            (<p>(.*\n)*?.*?</p>)
            ''', re.X | re.M)
        text = para_re.sub(self._newline_to_space_sub, text)
        text = self._unhash_code(text)
        return text

    _hash_code_dictionary = {}

    def _hash_code_sub(self, match):
        text = match.group(1)
        code_key = _hash_text(text)
        self._hash_code_dictionary[code_key] = text
        return code_key

    def _hash_code(self, text):
        code_re = re.compile(r'''
            (<pre><codeblock (.*\n)*?.*?</codeblock></pre>)
            ''', re.X | re.M)
        text = code_re.sub(self._hash_code_sub, text)
        return text

    def _unhash_code(self, text):
        for hash_key in self._hash_code_dictionary.keys():
            hash_re = re.compile('''%s''' % hash_key, re.X | re.M)
            text = hash_re.sub(self._hash_code_dictionary[hash_key], text)
        return text

    def _delete_newline(self, text):
        excessive_newline_re = re.compile(r'''
            \n{2,}
            ''', re.X | re.M)
        text = excessive_newline_re.sub('\n', text)
        return text

    def _do_document(self, text):
        setdocument_re = re.compile(r'''
            (\A)''', re.X | re.M)
        text = setdocument_re.sub('<document>\n', text)
        setdocument_re = re.compile(r'''
            (\Z)''', re.X | re.M)
        text = setdocument_re.sub('</document>', text)
        return text

    _block_quote_re = re.compile(r'''
        (                           # Wrap whole match in \1
        (
        ^[ \t]*>[ \t]?          # '>' at the start of a line
        .+\n                  # rest of the first line
        (.+\n)*                 # subsequent consecutive lines
        \n*                     # blanks
        )+
        )
        ''', re.M | re.X)

    _bq_one_level_re = re.compile('^[ \t]*>[ \t]?', re.M);

    _html_pre_block_re = re.compile(r'(\s*<pre>.+?</pre>)', re.S)
    def _dedent_two_spaces_sub(self, match):
        return re.sub(r'(?m)^  ', '', match.group(1))

    def _block_quote_sub(self, match):
        bq = match.group(1)
        bq = self._bq_one_level_re.sub('', bq)  # trim one level of quoting
        bq = self._ws_only_line_re.sub('', bq)  # trim whitespace-only lines
        bq = self._run_block_gamut(bq)          # recurse
        bq+="\n"
        #insert end of title
        bq = self._header.form_title(bq)
        bq = re.sub('(?m)^', '  ', bq)
        # These leading spaces screw with <pre> content, so we need to fix that:
        bq = self._html_pre_block_re.sub(self._dedent_two_spaces_sub, bq)

        return "<quote>\n%s\n</quote>\n\n" % bq

    def _do_block_quotes(self, text):
        if '>' not in text:
            return text
        return self._block_quote_re.sub(self._block_quote_sub, text)

    def _form_paragraphs(self, text):
        # Strip leading and trailing lines:
        text = text.strip('\n')
        # Wrap <p> tags.
        grafs = []
        for i, graf in enumerate(re.split(r"\n{2,}", text)):
            if graf in self._hash_html.get_htmlblocks_hash():
                # Unhashify HTML blocks
                html_bl = self._hash_html.get_htmlblocks_hash()
                grafs.append(html_bl[graf])
            else:
                cuddled_list = None
                if "cuddled-lists" in self.extras:
                    # Need to put back trailing '\n' for `_list_item_re`
                    # match at the end of the paragraph.
                    li = self._list_item_re.search(graf + '\n')
                    # Two of the same list marker in this paragraph: a likely
                    # candidate for a list cuddled to preceding paragraph
                    # text (issue 33). Note the `[-1]` is a quick way to
                    # consider numeric bullets (e.g. "1." and "2.") to be
                    # equal.
                    if (li and len(li.group(2)) <= 3 and li.group("next_marker")
                        and li.group("marker")[-1] == li.group("next_marker")[-1]):
                        start = li.start()
                        cuddled_list = self._list.do_lists(graf[start:]).rstrip("\n")
                        assert cuddled_list.startswith("<list")
                        graf = graf[:start]

                # Wrap <p> tags.<strong><emphasis> tags
                graf = self._spangamut.run_span_gamut(graf)
                grafs.append("<p>" + graf.lstrip(" \t") + "</p>")
                if cuddled_list:
                    grafs.append(cuddled_list)

        return "\n\n".join(grafs)

    def _add_footnotes(self, text):
        if self.footnotes:
            footer = [
                      '<div class="footnotes">',
                      '<hr' + self.empty_element_suffix,
                      '<ol>',
                      ]
            for i, id in enumerate(self.footnote_ids):
                if i != 0:
                    footer.append('')
                footer.append('<li id="fn-%s">' % id)
                footer.append(self._run_block_gamut(self.footnotes[id]))
                backlink = ('<a href="#fnref-%s" '
                            'class="footnoteBackLink" '
                            'title="Jump back to footnote %d in the text.">'
                            '&#8617;</a>' % (id, i + 1))
                if footer[-1].endswith("</p>"):
                    footer[-1] = footer[-1][:-len("</p>")] \
                        + '&nbsp;' + backlink + "</p>"
                else:
                    footer.append("\n<p>%s</p>" % backlink)
                footer.append('</li>')
            footer.append('</ol>')
            footer.append('</div>')
            return text + '\n\n' + '\n'.join(footer)
        else:
            return text

    def _unescape_special_chars(self, text):
        # Swap back in all the special characters we've hidden.
        for ch, hash in self._escape_table.items():
            text = text.replace(hash, ch)
        return text

    def _outdent(self, text):
        # Remove one level of line-leading tabs or spaces
        return self._outdent_re.sub('', text)


# Recipe: dedent (0.1.2)
def _dedentlines(lines, tabsize=8, skip_first_line=False):
    """_dedentlines(lines, tabsize=8, skip_first_line=False) -> dedented lines

        "lines" is a list of lines to dedent.
        "tabsize" is the tab width to use for indent width calculations.
        "skip_first_line" is a boolean indicating if the first line should
            be skipped for calculating the indent width and for dedenting.
            This is sometimes useful for docstrings and similar.

    Same as dedent() except operates on a sequence of lines. Note: the
    lines list is modified **in-place**.
    """
    DEBUG = False
    if DEBUG:
        print "dedent: dedent(..., tabsize=%d, skip_first_line=%r)"\
              % (tabsize, skip_first_line)
#    indents = []
    margin = None
    for i, line in enumerate(lines):
        if i == 0 and skip_first_line: continue
        indent = 0
        for ch in line:
            if ch == ' ':
                indent += 1
            elif ch == '\t':
                indent += tabsize - (indent % tabsize)
            elif ch in '\r\n':
                continue  # skip all-whitespace lines
            else:
                break
        else:
            continue  # skip all-whitespace lines
        if DEBUG: print "dedent: indent=%d: %r" % (indent, line)
        if margin is None:
            margin = indent
        else:
            margin = min(margin, indent)
    if DEBUG: print "dedent: margin=%r" % margin

    if margin is not None and margin > 0:
        for i, line in enumerate(lines):
            if i == 0 and skip_first_line: continue
            removed = 0
            for j, ch in enumerate(line):
                if ch == ' ':
                    removed += 1
                elif ch == '\t':
                    removed += tabsize - (removed % tabsize)
                elif ch in '\r\n':
                    if DEBUG: print "dedent: %r: EOL -> strip up to EOL" % line
                    lines[i] = lines[i][j:]
                    break
                else:
                    raise ValueError("unexpected non-whitespace char %r in "
                                     "line %r while removing %d-space margin"
                                     % (ch, line, margin))
                if DEBUG:
                    print "dedent: %r: %r -> removed %d/%d"\
                          % (line, ch, removed, margin)
                if removed == margin:
                    lines[i] = lines[i][j + 1:]
                    break
                elif removed > margin:
                    lines[i] = ' ' * (removed - margin) + lines[i][j + 1:]
                    break
            else:
                if removed:
                    lines[i] = lines[i][removed:]
    return lines


def _dedent(text, tabsize=8, skip_first_line=False):
    """_dedent(text, tabsize=8, skip_first_line=False) -> dedented text

        "text" is the text to dedent.
        "tabsize" is the tab width to use for indent width calculations.
        "skip_first_line" is a boolean indicating if the first line should
            be skipped for calculating the indent width and for dedenting.
            This is sometimes useful for docstrings and similar.

    textwrap.dedent(s), but don't expand tabs to spaces
    """
    lines = text.splitlines(1)
    _dedentlines(lines, tabsize=tabsize, skip_first_line=skip_first_line)
    return ''.join(lines)
