#!/usr/bin/env python
# Copyright (C) 2013, Xcoo, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-

from mdparser import Markdown
import re
from lxml import etree
try:
    from hashlib import md5
except ImportError:
    from md5 import md5
from paragraph import Paragraph
from hashlist import HashList
from reader import Reader


class MdReader(Reader):

    def __init__(self, input_path, output_path):
        Reader.__init__(self, input_path, output_path)
        self._convert_dictionary = {}
        self._input_path = unicode(input_path, 'utf-8')

    def parse_md(self):
        self._hash = HashList()
        self._para = Paragraph()
        self._hash_em = HashList()
        
        markdowner = Markdown()
        md = markdowner.convert(self.get_buf())
        result = self._convert_heavy_add_path(md.encode('utf-8'))
        result = self._hash_emp(result)
        result = self._convert_items(result)
        result = self._hash_em.un_hash(result)
        result = self._delete_p_tag(result)
        #result = self._convert_ampsharp(result)
        result = self._convert_codeblock(result)
        result = self._convert_html_tag(result)
        return result
    
    def _convert_codeblock(self, text):
        root = etree.XML(text)
        pres = root.xpath('//pre')
        for pre in pres:
            code = pre[0]
            pre.clear()
            pre.tag = code.tag
            pre.text = code.text
            at = code.attrib
            for a in at:
                    pre.set(a, at[a])
        codes = root.xpath('//codeblock')
        for code in codes:
            code.text = etree.CDATA(code.text)
        return etree.tostring(root, pretty_print=True, encoding='utf-8')
    
    def _convert_ampsharp(self,text):
        amps_re = re.compile(r'''&amp;#''', re.X | re.M)
        text = amps_re = amps_re.sub('''&#''', text)
        return text
    
    def _delete_p_tag(self, text):
        p_re = re.compile(r'''<p/>\n</document''', re.X | re.M)
        text = p_re.sub("</document", text)
        return text
    
    def _strike_sub(self, match):
        content = match.group(1)
        return "<strike>%s</strike>" % content
    
    def _convert_html_tag(self, text):
        strike_re = re.compile("&lt;strike&gt;(.*?)&lt;/strike&gt;", re.X | re.M)
        text = strike_re.sub(self._strike_sub, text)
        strike_re = re.compile("&lt;s&gt;(.*?)&lt;/s&gt;", re.X | re.M)
        text = strike_re.sub(self._strike_sub, text)
        return text

    def _hash_emp(self, text):
        em_re = re.compile(r'''(<emphasis>.*?</emphasis>)''', re.X | re.M)
        text = self._hash_em.set_e_hash(text, em_re)
        strong_re = re.compile(r'''(<strong>.*?</strong>)''', re.X | re.M)
        text = self._hash_em.set_e_hash(text, strong_re)
        heavy_re = re.compile(r'''(<heavy>.*?</heavy>)''', re.X | re.M)
        text = self._hash_em.set_e_hash(text, heavy_re)
        code_re = re.compile(r'''(<codeblock .*?</code>)''', re.X | re.M)
        text = self._hash_em.set_e_hash(text, code_re)
        link_re = re.compile(r'''(<link .*?</link>)''', re.X | re.M)
        text = self._hash_em.set_e_hash(text, link_re)
        image_re = re.compile(r'''(<image .*?/>)''', re.X | re.M)
        text = self._hash_em.set_e_hash(text, image_re)
        return text

    def _convert_items(self, text):
        roots = etree.XML(text)
        for child in roots:
            self._list_children(child)
        result = self._hash.un_hash_with_tag(etree.tostring(roots, pretty_print=True, encoding = 'utf-8'), "<item>")
        return result

    def _convert_heavy_add_path(self, text):
        roots = etree.XML(text)
        roots.set("path", self._input_path)
        for child in roots:
            self._check_heavy(child)
        result = etree.tostring(roots, pretty_print = True, encoding = 'utf-8')
        for convert_key in self._convert_dictionary:
            convert_re = re.compile('''%s''' % convert_key, re.X | re.M)
            result = convert_re.sub(self._convert_dictionary[convert_key], result)
        return result

    _item_tag = "<item>|</item>"
    _item_head_re = re.compile(r'''(^item\|)''', re.X | re.M)
    _number_head_re = re.compile(r'''(^number\|)''', re.X | re.M)

    def _list_children(self, parent):
        if len(parent) != 0:
            index = 0
            num_or_normal = True
            for children in parent:
                self._list_children(children)
                if index == 0:
                    if parent.get('type') == "number":
                        num_or_normal = True
                    else:
                        num_or_normal = False

                if children.tag == "item":
                    text = self._para.get_content(children, self._item_tag)
                    if num_or_normal:
                        if re.search(self._number_head_re, text):
                            text = self._number_head_re.sub("", text)
                            text = "<item>" + text
                            text = self._hash.create_hash(text)
                            num_or_normal = True
                        if re.search(self._item_head_re, text):
                            text = self._item_head_re.sub("", text)
                            text = "</list>\n<list type = \"normal\">\n<item>" + text
                            text = self._hash.create_hash(text)
                            num_or_normal = False
                    else:
                        if re.search(self._item_head_re, text):
                            text = self._item_head_re.sub("", text)
                            text = "<item>" + text
                            text = self._hash.create_hash(text)
                            num_or_normal = False
                        if re.search(self._number_head_re, text):
                            text = self._number_head_re.sub("", text)
                            text = "</list>\n<list type = \"number\">\n<item>" + text
                            text = self._hash.create_hash(text)
                            num_or_normal = True
                    children.text = text
                index += 1

    _heavy_re = re.compile(r'''<heavy>''', re.X | re.M)
    _heavyend_re = re.compile(r'''</heavy>''',re.X | re.M)
                    
    def _refactor_heavy(self, tag, text):
        result = self._heavy_re.sub("</%s><heavy>" % tag, text)
        result = self._heavyend_re.sub("</heavy><%s>" % tag, result)
        self._convert_dictionary[text] = result
        result = "<tmpformat_heavy>"+result+"</tmpformat_heavy>"
        return etree.XML(result)
        
                    
    def _check_heavy(self, parent):
        if len(parent) != 0:
            for children in parent:
                self._check_heavy(children)
                if parent.tag == "strong" and children.tag == "emphasis":
                    children.tag = "heavy"
                    parent = self._refactor_heavy(parent.tag,  etree.tostring(parent, encoding = 'utf-8'))
                if parent.tag == "emphasis" and children.tag == "strong":
                    children.tag = "heavy"
                    parent = self._refactor_heavy(parent.tag,  etree.tostring(parent, encoding = 'utf-8'))
