#!/usr/bin/env python
# Copyright (C) 2013, Xcoo, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-

import re
try:
    from hashlib import md5
except ImportError:
    from md5 import md5
from links import Links
from hashhtml import HashHtml

try:
    import uuid
except ImportError:
    SECRET_SALT = str(randint(0, 1000000))
else:
    SECRET_SALT = str(uuid.uuid4())

def _hash_link(s):
    return 'md5' + md5(SECRET_SALT + s.encode("utf-8")).hexdigest()

class SpanGamut(object):

    def __init__(self, escapetable, ex, empty_element, code, tab_width):
        self._escape_table = escapetable
        self.extras = ex
        self._links = Links(ex, escapetable, empty_element, tab_width)
        self.empty_element_suffix = empty_element
        self._code = code
        self._hash_link_dictionary = {}
        self._links_re = re.compile('''(<link.*?</link>)''' , re.X | re.M)
        self._image_re = re.compile('''(<image.*?/>)''', re.X | re.M)

    # Ampersand-encoding based entirely on Nat Irons's Amputator MT plugin:
    #   http://bumppo.net/projects/amputator/
    #_ampersand_re = re.compile(r'&(?!#?[xX]?(?:[0-9a-fA-F]+|\w+);)')
    _ampersand_re = re.compile(r'&')
    #TEST 
    #_naked_lt_re = re.compile(r'<(?![a-z/?\$!])', re.I)
    #_naked_gt_re = re.compile(r'''(?<![a-z0-9?!/'"-])>''', re.I)
    _naked_lt_re = re.compile(r'<')
    _naked_gt_re = re.compile(r'''>''')
    
    def convert_amps_and_angles(self, text):
        # Smart processing for ampersands and angle brackets that need
        # to be encoded.
        text = self._ampersand_re.sub('&amp;', text)
        
        # Encode naked <'s
        text = self._naked_lt_re.sub('&lt;', text)
        
        # Encode naked >'s
        # Note: Other markdown implementations (e.g. Markdown.pl, PHP
        # Markdown) don't do this.
        text = self._naked_gt_re.sub('&gt;', text)
        return text

    def _encode_amps_and_angles(self, text):
        # Smart processing for ampersands and angle brackets that need
        # to be encoded.
        #text = self._ampersand_re.sub('&amp;', text)

        # Encode naked <'s
        #text = self._naked_lt_re.sub('&lt;', text)

        # Encode naked >'s
        # Note: Other markdown implementations (e.g. Markdown.pl, PHP
        # Markdown) don't do this.
        #text = self._naked_gt_re.sub('&gt;', text)
        return text

    def _encode_backslash_escapes(self, text):
        for ch, escape in self._escape_table.items():
            text = text.replace("\\" + ch, escape)
        return text

    # "Sorta" because auto-links are identified as "tag" tokens.
    _sorta_html_tokenize_re = re.compile(r"""
        (
        # tag
        </?
        (?:\w+)                                     # tag name
        (?:\s+(?:[\w-]+:)?[\w-]+=(?:".*?"|'.*?'))*  # attributes
        \s*/?>
        |
        # auto-link (e.g., <http://www.activestate.com/>)
        <\w+[^>]*>
        |
        <!--.*?-->      # comment
        |
        <\?.*?\?>       # processing instruction
        )
        """, re.X)

    def _escape_special_chars(self, text):
        # Python markdown note: the HTML tokenization here differs from
        # that in Markdown.pl, hence the behaviour for subtle cases can
        # differ (I believe the tokenizer here does a better job because
        # it isn't susceptible to unmatched '<' and '>' in HTML tags).
        # Note, however, that '>' is not allowed in an auto-link URL
        # here.
        escaped = []
        is_html_markup = False
        for token in self._sorta_html_tokenize_re.split(text):
            if is_html_markup:
                # Within tags/HTML-comments/auto-links, encode * and _
                # so they don't conflict with their use in Markdown for
                # italics and strong.  We're replacing each such
                # character with its corresponding MD5 checksum value;
                # this is likely overkill, but it should prevent us from
                # colliding with the escape values by accident.
                escaped.append(token.replace('*', self._escape_table['*'])
                               .replace('_', self._escape_table['_']))
            else:
                escaped.append(self._encode_backslash_escapes(token))
            is_html_markup = not is_html_markup
        return ''.join(escaped)
    
    def _hash_link_content(self, match):
        l_key = _hash_link(match.group(1))
        self._hash_link_dictionary[l_key] = match.group(1)
        return l_key

    def _set_link_hash(self, text):
        text = self._links_re.sub(self._hash_link_content, text)
        text = self._image_re.sub(self._hash_link_content, text)
        return text
        
            
            
    def _unhash_link(self,text):
        for hash_key in self._hash_link_dictionary.keys():
            hash_re = re.compile('''%s''' % hash_key, re.X | re.M)
            text = hash_re.sub(self._hash_link_dictionary[hash_key], text)
        return text

    def run_span_gamut(self, text):
        # These are all the transformations that occur *within* block-level
        # tags like paragraphs, headers, and list items.

        text = self._code.do_code_spans(text)
        text = self._escape_special_chars(text)
        
        # Process anchor and image tags.
        text = self._links.do_links(text)
        # Make links out of things like `<http://example.com/>`
        # Must come after _do_links(), because you can use < and >
        # delimiters in inline links like [this](<url>).
        text = self._links.do_auto_links(text)
        if "link-patterns" in self.extras:
            text = self._links.do_link_patterns(text)
        
        

        text = self._encode_amps_and_angles(text)
                
        text = self._set_link_hash(text)

        text = self._do_italics_and_bold(text)
    

        if "smarty-pants" in self.extras:
            text = self._do_smart_punctuation(text)
                
        text = self._unhash_link(text)

        # Do hard breaks:
        text = re.sub(r" {2,}\n", " <br%s\n" % self.empty_element_suffix, text)

        return text

    #_strong_re = re.compile(r"(\*\*|__)(?=\S)(.+?[*_]*)(?<=\S)\1", re.S)
    _strong_re = re.compile(r"(\*\*|__)(?=\S)(.+?)(?<=\S)\1", re.S)
    _em_re = re.compile(r"(\*|_)(?=\S)(.+?)(?<=\S)\1", re.S)
    _code_friendly_strong_re = re.compile(r"\*\*(?=\S)(.+?[*_]*)(?<=\S)\*\*", re.S)
    _code_friendly_em_re = re.compile(r"\*(?=\S)(.+?)(?<=\S)\*", re.S)
    _code_friendly_strong_em_re = re.compile(r"(<strong><emphasis>(.+?)</emphasis></strong>)", re.S)
    _code_strong_em_re = re.compile(r"(<strong><emphasis>(.+?)</strong></emphasis>)", re.S)
    _code_refactor_em_re = re.compile(r"<strong></emphasis>",re.S)

    def _do_italics_and_bold(self, text):
        # <strong> must go first:
        if "code-friendly" in self.extras:
            text = self._code_friendly_strong_re.sub(r"<strong>\1</strong>", text)
            text = self._code_friendly_em_re.sub(r"<emphasis>\1</emphasis>", text)
        else:
            text = self._strong_re.sub(r"<strong>\2</strong>", text)
            text = self._em_re.sub(r"<emphasis>\2</emphasis>", text)

        text = self._code_refactor_em_re.sub(r"</emphasis><strong>",text)
        text = self._code_strong_em_re.sub(r"<heavy>\2</heavy>", text)
        return text

    # "smarty-pants" extra: Very liberal in interpreting a single prime as an
    # apostrophe; e.g. ignores the fact that "round", "bout", "twer", and
    # "twixt" can be written without an initial apostrophe. This is fine because
    # using scare quotes (single quotation marks) is rare.
    _apostrophe_year_re = re.compile(r"'(\d\d)(?=(\s|,|;|\.|\?|!|$))")
    _contractions = ["tis", "twas", "twer", "neath", "o", "n",
                     "round", "bout", "twixt", "nuff", "fraid", "sup"]

    def _do_smart_contractions(self, text):
        text = self._apostrophe_year_re.sub(r"&#8217;\1", text)
        for c in self._contractions:
            text = text.replace("'%s" % c, "&#8217;%s" % c)
            text = text.replace("'%s" % c.capitalize(),
                                "&#8217;%s" % c.capitalize())
        return text

    # Substitute double-quotes before single-quotes.
    _opening_single_quote_re = re.compile(r"(?<!\S)'(?=\S)")
    _opening_double_quote_re = re.compile(r'(?<!\S)"(?=\S)')
    _closing_single_quote_re = re.compile(r"(?<=\S)'")
    _closing_double_quote_re = re.compile(r'(?<=\S)"(?=(\s|,|;|\.|\?|!|$))')

    def _do_smart_punctuation(self, text):
        """Fancifies 'single quotes', "double quotes", and apostrophes.
            Converts --, ---, and ... into en dashes, em dashes, and ellipses.
            Inspiration is: <http://daringfireball.net/projects/smartypants/>
            See "test/tm-cases/smarty_pants.text" for a full discussion of the
            support here and
            <http://code.google.com/p/python-markdown2/issues/detail?id=42> for a
            discussion of some diversion from the original SmartyPants.
            """
        if "'" in text:  # guard for perf
            text = self._do_smart_contractions(text)
            text = self._opening_single_quote_re.sub("&#8216;", text)
            text = self._closing_single_quote_re.sub("&#8217;", text)

        if '"' in text:  # guard for perf
            text = self._opening_double_quote_re.sub("&#8220;", text)
            text = self._closing_double_quote_re.sub("&#8221;", text)

        text = text.replace("---", "&#8212;")
        text = text.replace("--", "&#8211;")
        text = text.replace("...", "&#8230;")
        text = text.replace(" . . . ", "&#8230;")
        text = text.replace(". . .", "&#8230;")
        return text
