#!/usr/bin/env python
# Copyright (C) 2013, Xcoo, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-

__author__ = "Y.Ban"

import re
from writer import Writer
from hashlist import HashList
from paragraph import Paragraph
from lxml import etree
try:
    from hashlib import md5
except ImportError:
    from md5 import md5


class MdWriter(Writer):
    """Class for converting xml to markdown"""
    def __init__(self, xmlpath, outputpath):
        Writer.__init__(self, xmlpath, outputpath)
    
    def _add_line(self, child):
        text = ""
        if child.tag == "line":
            text = "\n---\n\n"
        return text

    def _set_title(self, child, quote_num):
        text = ""
        cnt = self._para.get_title_num(child.tag)
        if cnt > 0:
            text = ">" * quote_num + "#" * cnt + " " + child.get('title').encode('utf-8') + "\n\n"
        return text

    def _add_list_h(self, child, text, item_num):
        parent = child.getparent()
        list_h = ""
        if parent.get('type') == "normal":
            list_h = "- "
        elif parent.get('type') == "number":
            list_h = "%d. " % item_num
        text = list_h + text
        return text

    def _add_em_strong(self, text, child):
        for em in list(child):
            em_re = re.compile('''</?%s>''' % em.tag, re.X | re.M)
            if not self._para.is_html_tag(em.tag):
                text = em_re.sub("*" * self._para.get_em_num(em.tag), text)
            if len(em) != 0:
                text = self._add_em_strong(text, em)
        return text

    def _convert_space2newline(self, text):
        # Hash for equal and hyphen.
        equal_re = re.compile(r'''([ \t]+=+($|[ \t]+))''', re.X | re.M)
        text = self._hash.set_e_hash(text, equal_re)
        hyphen_re = re.compile(r'''([ \t]+-+($|[ \t]+))''', re.X | re.M)
        text = self._hash.set_e_hash(text, hyphen_re)
        # Convert space to new line.
        space_re = re.compile(r'''([ \t]+)''', re.X | re.M)
        text = space_re.sub("\n", text)
        return text

    def _add_quote(self, child, text, quote_num):
        text = ">" * quote_num + text
        return text

    def _set_paragraph(self, parent, item_num, quote_num):
        p_or_item = "<p>|</p>|<item>|</item>"
        text = self._para.get_content(parent, p_or_item)
        if re.search('<codeblock>', etree.tostring(parent, encoding = 'utf-8')) == None:
            children = self._hash.set_hash(text, parent)
            children = self._para.add_escape(children)
            text = self._add_em_strong(self._para.get_content(children, p_or_item), children)
            text = self._hash.un_hash(text)
        else:
            code_re = re.compile(r'''(<(/?codeblock>))''', re.X | re.M)
            text = code_re.sub("`", text)
        text += "\n"
        if parent.tag == "p":
            text += "\n"
        elif parent.tag == "item":
            text = self._add_list_h(parent, text, item_num)
        text = self._add_quote(parent, text, quote_num)
        text = self._para.convert_entity(text)
        return text

    def _set_code(self, parent):
        code_tag = "<codeblock>|</codeblock>"
        language = ""
        is_git = self._gitcode
        if parent.get('class') is not None:
            if parent.get('class') == "git":
                is_git = True
        if parent.get('language') is not None:
            language = parent.get('language')
            text = self._para.get_content(parent, '''<codeblock language="%s">|</codeblock>''' % language)
        else:
            text = self._para.get_content(parent, code_tag)

        text = self._para.convert_entity(text)
                #newline_re = re.compile(r'\n', re.X | re.M)
                #splitedline = newline_re.split(text)
        splitedline = text.splitlines()
        if language != "":
            if is_git:
                result = '''```%s\n''' % language
            else:
                result = '''<!--%s-->\n\n''' % language
        else:
            result = ""
        for line in splitedline:
            if is_git:
                result += line + "\n"
            else:
                result += "\t" + line + "\n"

        if is_git:
            result += "```\n\n"
        return result

    def xml2md(self, root, gitcode=False):
        """Convert xml to markdown."""
        self._hash = HashList()
        self._para = Paragraph()
        self._gitcode = gitcode
        text = ""
        for child in root:
            text += self._list_children(0, 0, 0, child)
            if child.tag == "list":
                text += "\n"
        text = self.unhash_html(text)
        return text

    def _list_children(self, indent_num, item_num, quote_num, parent):
        text = ""
        if parent.tag == "quote":
            quote_num += 1
        text += self._set_title(parent, quote_num)
        # for children in parent:
        if parent.tag == "hash":
            text += "\n"+parent.text+"\n\n"
        if parent.tag == "p" or parent.tag == "item":
            text += "\t" * indent_num + self._set_paragraph(parent, item_num, quote_num)
        elif parent.tag == "codeblock":
            text += self._set_code(parent)
        elif len(parent) != 0:
            item_num = 0
            for children in parent:
                etree.tostring(children, encoding='utf-8')
                if parent.get('type') == "number":
                    if children.tag == "item":
                        item_num += 1
                if parent.tag == "list" and children.tag == "list":
                    text += self._list_children(indent_num + 1, item_num, quote_num, children)
                else:
                    text += self._list_children(indent_num, item_num, quote_num, children)
                if children.tag == "list" and parent.tag != "list":
                    text += "\n"
        else:
            text += self._add_line(parent)
        if isinstance(text, unicode):
            text = text.encode("utf-8")
        return text
